// Text format: Line width 72, Ident 4
// CIF DRIVER for kernel 2.6.18
/*
   Common include files. They need to make a kernel module
*/
#include <linux/module.h>		
#include <linux/init.h>
#include <linux/kernel.h>		/* printk() 	*/
#include <linux/slab.h>			/* kzalloc() 	*/
#include <linux/errno.h>		/* error codes 	*/
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/wait.h>
#include <linux/interrupt.h>
#include <asm/system.h>			/* cli(), *_flags */
#include <asm/uaccess.h>
#include <asm/io.h>

/*
   Iclude file special for CIF device driver
*/
#include "../inc/cif_types.h"
#include "../inc/cif_dev_i.h"  	/* local definitions */
#include "../inc/cif_dev.h"   	/* local definitions */

/*
   This PCMCIA module use CIF_DEBUG to control debugging.  You can 
   define this variable in Makefile. Check the Makefile how to do it.
   If you do not define CIF_DEBUG at all, all the debug code will be
   left out. You can find debug macros in ../inc/cif_types.h
*/
#define DEVICE_NAME		"cif"	
#define MODULE_NAME		"cif_cs"
#define VERSION			"2.618"

/*********************************************************************/
/* External prototypes                                               */
/*********************************************************************/

extern  int              cif_ioctl         ( struct inode *, 
        struct file *, unsigned int , unsigned long ); 
extern  irqreturn_t      cif_interrupt     ( int , void *, 
        struct pt_regs *);
extern  void             cif_set_irq_state ( unsigned short, 
        DEV_INSTANCE *);
extern  void             cif_test_hardware ( DEV_INSTANCE *);
extern  DEV_INSTANCE    *cif_dev_get_at    ( int n); 

/*********************************************************************/
/* Global variables                                                  */
/*********************************************************************/

/*
   Global pointer to the linked list of the structures DEV_INSTANCE
*/
DEV_INSTANCE *cif_devices; 

/*
   Driver version
*/
char * pabDrvVersion = "CIFDrv V2.618-PCMCIA";

/*
   Major number of the CIF device. CIF_MAJOR=0 means number is arbitrary
*/
int cif_major   = CIF_MAJOR;

/*
   Current number of the CIF devices. Initial value is 0 - no devices
*/
int cif_nr_devs = 0;

static struct class *cif_class;

/*********************************************************************/
/* Function prototypes                                               */
/*********************************************************************/

/*
   The cif60_config() and cif60_release() entry points are used 
   to configure or release a socket, in response to card insertion and
   ejection events.
*/
static  int  cif60_config ( struct pcmcia_device *link);
static void  cif60_release( struct pcmcia_device *link);

/*
   The cif60_attach() and cif60_detach() entry points are used to create
   and destroy "instances" of the driver - DEV_INSTANCE objects, where 
   each instance represents everything needed to manage one actual 
   PCMCIA card.
*/
static  int  cif60_attach( struct pcmcia_device *link);
static void  cif60_detach( struct pcmcia_device *link);

/*********************************************************************/
/* MODULE DOCUMENTATION                                              */
/*********************************************************************/

MODULE_AUTHOR("David Tsaava and Ivan Ibrin");
MODULE_DESCRIPTION("CIFDriver for Hilscher PCMCIA Communication Board");
MODULE_SUPPORTED_DEVICE("CIF PB-60"); 
MODULE_ALIAS ("cif"); 
MODULE_LICENSE("GPL and additional rights");
MODULE_VERSION (VERSION);

/*********************************************************************/
/* FILE OPERATIONS                                                   */
/*********************************************************************/

/*
   Almost all the following file operations are actually unuseful. 
   cif_ioctl() do all the data exchange operations.

   The llseek method is used to change the current read/write position 
   in a file, and the new position is returned as a (positive) return 
   value. The loff_t parameter is a “long offset” and is at least 64 
   bits wide even on 32-bit platforms. Errors are signaled by a negative
   return value. 
*/
static loff_t cif_llseek (struct file *filp, loff_t off, int whence)
{
	loff_t newpos;
	switch(whence) {
	case 0: /* SEEK_SET */
		newpos = off;
		break;
	case 1: /* SEEK_CUR */
		newpos = filp->f_pos + off;
		break;
	default: /* can't happen */
		return -EINVAL;
	}
	if (newpos<0) return -EINVAL;
	filp->f_pos = newpos;
	return newpos;
}

/*
   Used to retrieve data from the device. In our case it does nothing.
*/
static ssize_t cif_read (struct file *filp, char *buf, size_t count,
       loff_t *f_pos)
{
	return (0);
}
/*
   Sends data to the device. In our case it does nothing.
*/
static ssize_t cif_write (struct file *filp, const char *buf, 
	size_t count, loff_t *f_pos)
{
	return (-ENOMEM);
}

/*
   Though this is always the first operation performed on the device 
   file. 
*/
static int cif_open (struct inode *inode, struct file *filp)
{
	int num  =  NUM(inode->i_rdev);
	DEV_INSTANCE  *dev;

	/* 
	   Type 0, check the device number 
	*/
	if (num >= cif_nr_devs) return -ENODEV;
	dev = cif_devices;
	
	/* 
		Use filp->private_data to point to the device data 
	*/
	filp->private_data = dev;
	return 0;
}

/*
   This operation is invoked when the file structure is being released. 
*/
static int cif_release (struct inode *inode, struct file *filp)
{
    return (0);
}

/* 
   File operation structure.
   Each field in the structure must point to the function in the driver 
   that implements a specific operation, or be left NULL for unsupported
   operations. The following list introduces all the operations that an 
   application can invoke on a device.
*/
static struct file_operations cif_fops = {
	.owner      = THIS_MODULE,
	.llseek     = cif_llseek,
	.read       = cif_read,
	.write      = cif_write,
	.ioctl      = cif_ioctl, 
	.open       = cif_open,
	.release    = cif_release,
};

/*********************************************************************/
/* PCMCIA ROUTINES                                                   */
/*********************************************************************/

/*
   cif60_attach() creates an "instance" of the driver, allocating local 
   data structures for one device. We don't actually configure the card 
   at this point - we wait until we receive a card insertion event.
*/
static int cif60_attach(struct pcmcia_device *link)
{
	DEV_INSTANCE	*dev;
	//dev_link_t 		*link;

	cif_devices = (DEV_INSTANCE *)
	              kzalloc(sizeof(DEV_INSTANCE), GFP_KERNEL); 
	if (!cif_devices) return -ENOMEM;
	dev = (DEV_INSTANCE*)cif_devices; 
	dev->next = NULL;
	cif_nr_devs = 1;

	//link = &dev->link;
 	link->priv = dev; 
	dev->p_dev = link;
	
	//link->handle = p_dev;
	//p_dev->instance = link;
	DBG_PRN( "*** ATTACH (0x%p) ***\n", link);
	
	/* 
	   Entries the device structure 
	*/
	dev->usBoard        = 0;
	dev->ucPCIBusNumber = 0;//FIXME: INCOMPLETE CODE //link->handle->device_no; // Not used anywhere
	dev->bActive        = TRUE;
	dev->ucBusType      = BUS_TYPE_PCI; // Also we use PCMCIA ;)

	/* 
	   Do configuration of the socket
	*/
	//link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	cif60_config(link);

	/* 
	   At this moment the CIF card is fully configured. The last step 
	   is creating the device with node /dev/cif.
	   Because in MKDEV(cif_major, 0) cif_major=CIF_MAJOR=0, the major 
	   number of the device is arbitrary and minor number equals to 0.
	*/
	class_device_create(cif_class, NULL, MKDEV(cif_major, 0),
	                    NULL, "cif");
	DBG_PRN( "device %d:%d attached\n", cif_major, 0);
	return 0;
}

/* 
   This deletes a driver "instance". If it has been released, all 
   local data structures are freed.  Otherwise, the structures will 
   be freed when the device is released.
*/
static void cif60_detach(struct pcmcia_device *link)
{
	//dev_link_t *link = dev_to_instance(p_dev); // ds.h:197
	DEV_INSTANCE *dev = link->priv;
	
	DBG_PRN( "*** DETACH (0x%p) ***\n", link);
	//link->state &= ~DEV_PRESENT;
	/*if (link->state & DEV_CONFIG)*/ cif60_release(link);

	kfree(dev); 
	cif_nr_devs = 0; 
	DBG_PRN( "cif_nr_devs = %d\n", cif_nr_devs);

	class_device_destroy(cif_class, MKDEV(cif_major, 0));
	DBG_PRN( "device %d:%d detached\n", cif_major, 0);
	return;
}

/*
   Simple macro for checking the return values of the executing 
   functions and if return value is not equal to CS_SUCCESS (zero) goto
   cs_failed label.
*/
#define CS_CHECK(fn, ret) 										\
	do { fail_fn = (fn); 										\
		if ((fail_rc = (ret)) != CS_SUCCESS) goto cs_failed; 	\
} while (0)
#define MAX_TUPLE_SIZE 64

static int cif60_config(struct pcmcia_device *link)
{
	DEV_INSTANCE 	*dev = link->priv;
	//client_handle_t handle = link->handle; 
	tuple_t 		tuple;
	u_char 			buf[MAX_TUPLE_SIZE]; 
	cisparse_t 		parse;
	config_info_t 	config;
	cisinfo_t 		info;
	int 			fail_fn, fail_rc;
	win_req_t	 	req;
	memreq_t 		mem;

	DBG_PRN( "*** CONFIG (0x%p) ***\n", link);
	
	/*
	   Almost all the modern cards have CIS and we totally rely on it.
	   If the CIS validation passes we can begin configure the socket.
	*/	
	CS_CHECK(ValidateCIS, pcmcia_validate_cis(link, &info));

	/*
	   Tune up tuples 
	*/
	tuple.Attributes = 0; 
	tuple.TupleData = buf; 
	tuple.TupleDataMax = MAX_TUPLE_SIZE; 
	tuple.TupleOffset = 0;

	/* 
	   This reads the card's CONFIG tuple to find its configuration 
	   regsisters.
	*/
	tuple.DesiredTuple = CISTPL_CONFIG; 
	CS_CHECK (GetFirstTuple, pcmcia_get_first_tuple(link, &tuple));
	CS_CHECK (GetTupleData, pcmcia_get_tuple_data(link, &tuple));
	CS_CHECK (ParseTuple, pcmcia_parse_tuple(link, &tuple, &parse));
	link->conf.ConfigBase = parse.config.base;
	link->conf.Present = parse.config.rmask[0];

	/* 
	   Determine card type and firmware version. Here we don't use this
	   information but if we in doubt about the card identity we can
	   check it up.
	*/
	tuple.DesiredTuple = CISTPL_VERS_1; 
	CS_CHECK (GetFirstTuple, pcmcia_get_first_tuple(link, &tuple));
	CS_CHECK (GetTupleData, pcmcia_get_tuple_data(link, &tuple));
	CS_CHECK (ParseTuple, pcmcia_parse_tuple(link, &tuple, &parse));
	
	/* 
	   Determine device parameters. Here we find out the access
	   speed of the card common memory.
	*/
	tuple.DesiredTuple = CISTPL_DEVICE; 
	CS_CHECK (GetFirstTuple, pcmcia_get_first_tuple(link, &tuple));
	CS_CHECK (GetTupleData, pcmcia_get_tuple_data(link, &tuple));
	CS_CHECK (ParseTuple, pcmcia_parse_tuple(link, &tuple, &parse));
	req.AccessSpeed = parse.device.dev[0].speed;
	
	/* 
	   Determine the memory and io windows and their's parameters.
	   CIF device has only one memory window with size of 8 kBytes.
	*/
	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	CS_CHECK (GetFirstTuple, pcmcia_get_first_tuple(link, &tuple));
	CS_CHECK (GetTupleData, pcmcia_get_tuple_data(link, &tuple));
	CS_CHECK (ParseTuple, pcmcia_parse_tuple(link, &tuple, &parse));
	link->conf.ConfigIndex = parse.cftable_entry.index; 
	req.Base = parse.cftable_entry.mem.win[0].host_addr;
	req.Size = parse.cftable_entry.mem.win[0].len;
	mem.CardOffset = parse.cftable_entry.mem.win[0].card_addr;
	dev->ulDPMByteSize = parse.cftable_entry.mem.win[0].len; 
	DBG_PRN( "dev->ulDPMByteSize = %d\n", (int)dev->ulDPMByteSize);
	dev->ulDPMSize = dev->ulDPMByteSize / 1024;
	DBG_PRN( "dev->ulDPMSize = %d\n", (int)dev->ulDPMSize);

	/* 
	   Configure the card. Flip the configuration flag.
	*/
	//link->state |= DEV_CONFIG;

	/* 
	   Look up current Vcc 
	*/
	CS_CHECK (GetConfigurationInfo, 
		pcmcia_get_configuration_info(link, &config));
	//link->conf.Vcc = config.Vcc;
	DBG_PRN("config.Vcc = %d\n", config.Vcc);	

	/*
	   As CIF card use polling only the interrupts configuring section
	   is missing. Here we just set up the interrupt number in the 
	   device instance structure to zero. 
	*/
	dev->usBoardIrq = 0; // POLLING ONLY !
	DBG_PRN("dev->usBoardIrq = %d\n",dev->usBoardIrq);	

	dev->usBoardIrq_scanned = dev->usBoardIrq; 
	DBG_PRN("dev->usBoardIrq_scanned = %d\n",dev->usBoardIrq_scanned);	

	/*
	   General socket configuration defaults can go here. In this 
	   client, we assume very little, and rely on the CIS for almost 
	   everything. In most clients, many details (i.e., number, sizes, 
	   and attributes of IO windows) are fixed by the nature of the 
	   device, and can be hard-wired here.
	*/
	link->conf.Attributes	= 0;
	link->conf.IntType	 	= INT_MEMORY;
	link->conf.Present	 	= PRESENT_OPTION;

	/* 
	   This actually configures the PCMCIA socket.
	*/
	CS_CHECK(RequestConfiguration,
		pcmcia_request_configuration(link, &link->conf));

	/*
	   Set up window for common memory. Note that the memory window base
	   is a physical address, and needs to be mapped to virtual space 
	   with ioremap() before it is used.
	*/
	req.Attributes = WIN_DATA_WIDTH_8 | WIN_MEMORY_TYPE_CM | WIN_ENABLE;
	req.Attributes |= WIN_USE_WAIT;

	/* 
	   This reserves memory space but doesn't actually enable it 
	*/
	CS_CHECK(RequestWindow, 
		pcmcia_request_window(&link, &req, &link->win));
	DBG_PRN( "CM: Attr = %xh, Base = 0x%x, Size = %d, Speed = %d ns\n", 
		req.Attributes, (u_int)req.Base, req.Size, req.AccessSpeed);
	dev->ulBoardAddress = req.Base;
	DBG_PRN( "dev->ulBoardAddress = 0x%x\n", (int)dev->ulBoardAddress);

	/* 
	   This enable common memory page and remap it.
	*/
	mem.Page = 0; 
	CS_CHECK(MapMemPage, pcmcia_map_mem_page(link->win, &mem));
	dev->pDpmBase = (unsigned char *) ioremap(req.Base, req.Size);
	DBG_PRN( "dev->pDpmBase = 0x%p\n", dev->pDpmBase);

	dev->IrqCtrlRegAddr = NULL; // FIXME
	DBG_PRN( "dev->IrqCtrlRegAddr = 0x%p\n", dev->IrqCtrlRegAddr);

	/*
	   Get handle to message area, last Kbyte of the DPM
	*/
	dev->ptDpmAddress = (DPM_MEMORY *)
	                    (dev->pDpmBase + dev->ulDPMByteSize - 1024);
	DBG_PRN( "dev->ptDpmAddress = 0x%p\n",dev->ptDpmAddress);

	/*
	   Get handle to send data area (start of DPM)
	*/
	dev->pbDpmSendArea = dev->pDpmBase;
	DBG_PRN( "dev->pbDpmSendArea = 0x%p\n",dev->pbDpmSendArea);

	/*
	   Get handle to receive data area (middle of the DPM)
	*/
	dev->pbDpmReceiveArea = dev->pDpmBase + (dev->ulDPMByteSize-1024)/2;
	DBG_PRN( "dev->pbDpmReceiveArea = 0x%p\n", dev->pbDpmReceiveArea);

	/*
	   All parameters successfully read. Initialize the instance.
	*/
	dev->usRcsError    = 0;  // RCS-Error during startup
	dev->sDrvInitError = 0;  // DRV-Error during startup

	/*
	   Pointers to the DEV dpm cells
	*/
	dev->pbHostFlags = &(dev->ptDpmAddress->HostFlags);
	dev->pbDevFlags  = &(dev->ptDpmAddress->DevFlags);

	/*
	   Initialize start states
	*/
	dev->usOpenCounter  = 0;
	dev->bInitMsgFlag   = INI_MSG_WAIT;       // Init state
	dev->bReadMsgFlag   = HOST_MBX_EMPTY;     // Read state
	dev->bWriteMsgFlag  = DEVICE_MBX_EMPTY;   // write state
	dev->bExIOFlag      = EXCHGIO_NON;        // IO state
	dev->bStateFlag     = STATE_ERR_NON;      // State field

	init_waitqueue_head (&dev->pInitSemaphore);
	init_waitqueue_head (&dev->pReadSemaphore);
	init_waitqueue_head (&dev->pWriteSemaphore);
	init_waitqueue_head (&dev->pExIOSemaphore);
	spin_lock_init      (&dev->mutex);

	/*
	   At this point, the dev_node_t structure(s) need to be initialized
	   and arranged in a linked list at link->dev.
	*/
	//link->dev = &dev->node;
	//link->state &= ~DEV_CONFIG_PENDING;
	dev = link->priv;
	sprintf(dev->node.dev_name, DEVICE_NAME); // or strcpy
	dev->node.major = cif_major;
	dev->node.minor = 0; //devno;
	dev->node.next = NULL;

	/* 
	   Finally, report what we've done 
	*/
	DBG_PRN("V-IRQ                     = 0x%p\n",dev->pvVirtualIrq );
	DBG_PRN("V-DPM base address        = 0x%p\n",dev->pDpmBase );
	DBG_PRN("V-DPM send area address   = 0x%p\n",dev->pbDpmSendArea);
	DBG_PRN("V-DPM recive area address = 0x%p\n",dev->pbDpmReceiveArea);
	DBG_PRN("V-DPM last kByte          = 0x%p\n",dev->ptDpmAddress);

	//link->state &= ~DEV_CONFIG_PENDING;
	return 0;

	/*
	   In case of error the configuration sequence breaks and 
	   card releases.
	*/
cs_failed:
	DBG_PRN( "ERROR!!!\n");
	cs_error(link, fail_fn, fail_rc);
	cif60_release(link);
	return -ENODEV;
}

static void cif60_release(struct pcmcia_device *link)
{
	DEV_INSTANCE *dev = link->priv;

	DBG_PRN( "*** RELEASE (0x%p) ***\n", link);
	/* 
	   Don't bother checking to see if these succeed or not 
	*/
	//pcmcia_release_configuration(link->handle);
	pcmcia_disable_device(link);

	if (link->win) {
		iounmap(dev->pDpmBase);
		pcmcia_release_window(link->win);
	}

	//link->state &= ~DEV_CONFIG;
}


/*
   Card event - SUSPEND.
*/
/*
static int cif60_suspend(struct pcmcia_device *p_dev) 
{
	dev_link_t *link = dev_to_instance(p_dev);

	DBG_PRN( "*** SUSPEND (0x%p) ***\n", link);
	link->state |= DEV_SUSPEND;
	if (link->state & DEV_CONFIG)
		pcmcia_release_configuration(link->handle);
		// FIXME: Uncomplete code!
	return 0;
}
*/
/*
   Card event - RESUME.
*/
/*
static int cif60_resume(struct pcmcia_device *link)
{
	//dev_link_t *link = dev_to_instance(p_dev);

	DBG_PRN( "*** RESUME *** (0x%p) \n", link);
	link->state &= ~DEV_SUSPEND;
	if (link->state & DEV_CONFIG)
		pcmcia_request_configuration(link->handle, &link->conf);
		// FIXME: Uncomplete code! 
	return 0;
}
*/
/*
   Static structure to precisely indentify our CIF card
*/
static struct pcmcia_device_id cif60_ids[] = {
	PCMCIA_DEVICE_MANF_CARD(0x0042, 0x0042),
	PCMCIA_DEVICE_PROD_ID12("Hilscher GmbH", "CIF60 PB",
	                        0x448030a8, 0x6bee925a), PCMCIA_DEVICE_NULL,
};
MODULE_DEVICE_TABLE(pcmcia, cif60_ids);

static struct pcmcia_driver cif60_driver = {
  	.owner		= THIS_MODULE,
  	.drv		= {
		.name	= MODULE_NAME,
	},
	.probe		= cif60_attach,
	.remove		= cif60_detach,
	//.suspend	= cif60_suspend,
	//.resume		= cif60_resume,
	.id_table	= cif60_ids,
};

/*
   cif60_init() and cif60_exit() are the special functions, one to be 
   invoked when the module is loaded into kernel (cif60_init) and the 
   second one for when the module is removed. 
*/
static int __init cif60_init(void)
{
	int rc;
	printk(MODULE_NAME " loaded: [v. " VERSION "]\n");
	DBG_PRN("DEBUG is ON. special thanks to lesnikova_sveta@mail.ru\n");

	cif_class = class_create(THIS_MODULE, "cif_60"); 
	if (!cif_class) return -1;
	
	/* 
	   Register major, and accept a dynamic number 
	*/
	rc = register_chrdev(cif_major, DEVICE_NAME, &cif_fops);
	if (rc < 0) {
		DBG_PRN("can't get major %d\n", cif_major);
		return rc;
	}
	if (cif_major == 0) cif_major = rc; /* dynamic */
	DBG_PRN("major number is %d = %d\n", cif_major, rc);

	rc = pcmcia_register_driver(&cif60_driver);	
	if (rc < 0) {
		DBG_PRN("can't register driver, ret = %d\n", rc);
		return rc;
	}
	return 0;
}

static void __exit cif60_exit(void)
{
	pcmcia_unregister_driver(&cif60_driver);	
	unregister_chrdev(cif_major, DEVICE_NAME);	
	class_destroy(cif_class); 
	printk(MODULE_NAME " removed: [v. " VERSION "]\n");
}

/*
   module_init and module_exit lines use special kernel macros to 
   indicate the role of previous two functions. 
*/
module_init(cif60_init);
module_exit(cif60_exit);
